# MIT License Copyright (c) 2022 joh-fischer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os, sys
import logging
import datetime
import functools
from termcolor import colored

import yaml
import csv
import os
import torch
import datetime
from torch.utils.tensorboard import SummaryWriter


# command line & txt file logger

@functools.lru_cache()
def get_logger(log_dir='./logs', dist_rank=0, prefix=None):
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    timestamp = datetime.datetime.now().strftime('%y%m%d-%H%M%S')

    logger = logging.getLogger('')
    logger.setLevel(logging.INFO)
    logger.propagate = False

    # create formatter
    fmt = '[%(asctime)s] (%(filename)s %(lineno)d): %(levelname)s %(message)s'
    color_fmt = colored('[%(asctime)s]', 'green') + \
                colored('(%(filename)s %(lineno)d)', 'yellow') + \
                ': %(levelname)s %(message)s'

    # file writer
    prefix = prefix + "_" if prefix is not None else ""
    filename = f'log_{prefix}{timestamp}.txt'
    filepath = os.path.join(log_dir, filename)
    fh = logging.FileHandler(filepath)
    fh.setLevel(logging.INFO)
    fh.setFormatter(logging.Formatter(fmt=fmt, datefmt='%Y-%m-%d %H:%M:%S'))
    logger.addHandler(fh)

    # standard output (print), only on rank 0
    if dist_rank == 0:
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.INFO)
        sh.setFormatter(logging.Formatter(fmt=color_fmt, datefmt='%Y-%m-%d %H:%M:%S'))
        logger.addHandler(sh)

    return logger


# Logger class

class Aggregator:
    def __init__(self):
        """ Aggregate sum and average. """
        self.sum = 0.
        self.avg = 0.
        self.count = 0

    def update(self, val: float, n: int = 1):
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class Dummy:
    """ Mock class for tensorboard `SummaryWriter` """

    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        def func(*args, **kwargs):
            pass

        return func


class Logger:
    def __init__(self, log_dir: str = 'runs', exp_name: str = None, timestamp: bool = False,
                 tensorboard: bool = False):
        """
        Custom logger for PyTorch training loops.

        Notes
        ----------
        The logs are stored in <log_dir>/<exp_name>/<timestamp> if `exp_name` is set and `timestamp` is true. For
        example: './logs/model1/22-07-07_121028'. This folder then contains 'metrics.csv' and 'hparams.yaml'.

        Parameters
        ----------
        log_dir : str
            Base directory of experiment logs. Default: 'runs'.
        exp_name : str
            Experiment name to create specific folder. Logs are stored in <log_dir>/<exp_name>.
        timestamp : bool
            If true, makes a folder with the current time for the logs, e.g. <log_dir>/<exp_name>/<timestamp>. The
            timestamp format is `y-m-d_HMS`, e.g. 22-07-07_121028.
                tensorboard : bool
            If true, creates a `SummaryWriter` instance to write to tensorboard.
        """
        self.exp_name = exp_name if exp_name else ''
        self.timestamp = datetime.datetime.now().strftime('%y-%m-%d_%H%M%S') if timestamp else ''
        self.log_dir = os.path.join(log_dir, self.exp_name, self.timestamp)

        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir, exist_ok=True)

        self.writer = SummaryWriter(self.log_dir) if tensorboard else Dummy()

        self.metrics = []
        self.hparams = {}

        self.running_epoch = -1

        self.epoch = {}

    @property
    def tensorboard(self):
        """ Returns tensorboard `SummaryWriter` instance. """
        return self.writer

    def log_hparams(self, params: dict):
        """
        Log hyper-parameters.

        Parameters
        ----------
        params : dict
            Dictionary containing the hyperparameters as key-value pairs. For
            example {'optimizer': 'Adam', 'lr': 1e-02}.
        """
        self.hparams.update(params)

    def log_metrics(self, metrics_dict: dict, step: int = None, phase: str = '', aggregate: bool = False, n: int = 1):
        """
        Log metrics.

        Parameters
        ----------
        metrics_dict : dict
            Dictionary containing the metrics as key-value pairs. For example {'acc': 0.9, 'loss': 0.2}.
        step : int, optional
            Step number where metrics are to be recorded.
        phase : str, optional
            Current phase, e.g. 'train' or 'val' or 'epoch_end'.
        aggregate : bool
            If true, aggregates values into sum and average.
        n : int
            Count for aggregating values.
        """
        step = step if step is not None else len(self.metrics)

        metrics = {'epoch': self.running_epoch, 'phase': phase}
        metrics.update({k: self._handle_value(v) for k, v in metrics_dict.items()})
        metrics['step'] = step

        self.metrics.append(metrics)

        if aggregate:
            for metric_name, metric_val in metrics_dict.items():
                if metric_name not in self.epoch:
                    self.epoch[metric_name] = Aggregator()
                self.epoch[metric_name].update(metric_val, n)

    def init_epoch(self, epoch: int = None):
        """
        Initializes a new epoch.

        Parameters
        ----------
        epoch : int
            If empty, running epoch will be increased by 1.
        """
        if epoch:
            self.running_epoch = epoch
        else:
            self.running_epoch += 1

        self.epoch = {}

    def save(self, name: str = None):
        """
        Save the hyperparameters and metrics to a file.

        Parameters
        ----------
        name : str
            Additional name for saving the metrics and hyperparameters.
        """
        name = '_' + name if name else ''

        if self.metrics:
            metrics_file_path = os.path.join(self.log_dir, 'metrics' + name + '.csv')
            last_m = {}
            for m in self.metrics:
                last_m.update(m)
            metrics_keys = list(last_m.keys())

            with open(metrics_file_path, 'w') as file:
                writer = csv.DictWriter(file, fieldnames=metrics_keys)
                writer.writeheader()
                writer.writerows(self.metrics)

        if self.hparams:
            hparams_file_path = os.path.join(self.log_dir, 'hparams' + name + '.yaml')

            output = yaml.dump(self.hparams, Dumper=yaml.Dumper)
            with open(hparams_file_path, 'w') as file:
                file.write(output)

    @staticmethod
    def _handle_value(value):
        if isinstance(value, torch.Tensor):
            return value.item()
        return value


if __name__ == "__main__":
    l = get_logger('logs')
    l.info("Information")
    l.warning("Warning")



import torch
import torch.nn as nn
import einops


# TODO: add QKVAttention


class MultiHeadAttention(nn.Module):
    def __init__(self, input_dim: int, dim: int = 64, n_heads: int = 4):
        """
        Multi-Head Attention module as described in Attention Is All
        You Need" (https://arxiv.org/abs/1706.03762).
        Args:
            input_dim: Input dimension.
            dim: Dimension of keys, queries, values.
            n_heads: Number of heads for multi-head attention.
        """
        super().__init__()

        self.dim = dim

        self.to_qkv = nn.Linear(input_dim, dim * n_heads * 3, bias=False)
        self.scale = dim ** -0.5

        self.unify_heads = nn.Linear(dim * n_heads, input_dim)

    def forward(self, x: torch.Tensor):
        qkv = self.to_qkv(x)
        q, k, v = einops.rearrange(qkv, 'b t (qkv h d) -> qkv b h t d', qkv=3, d=self.dim)     # [bs, h, t, d]

        dot = torch.einsum('bhtd, bhkd -> bhtk', q, k) * self.scale     # [bs, h, t, t]

        att = torch.softmax(dot, dim=-1)

        res = torch.einsum('bhdt, bhtv -> bhdv', att, v)
        res = einops.rearrange(res, 'b h t d -> b t (h d)')
        res = self.unify_heads(res)

        return res


class ConvAttention(nn.Module):
    def __init__(self, n_channels: int, dim_keys: int = 32, n_heads: int = 2):
        """
        Applies self-attention like in "Attention Is All You Need" (https://arxiv.org/abs/1706.03762)
        to an image by reshaping it into a sequence. Only for small field sizes.

        Args:
            n_channels (int): Number of channels of the input feature maps
            dim_keys (int): Dimension of queries, keys, and values
            n_heads (int): Number of heads for attention
        """
        super().__init__()
        self.scale = dim_keys ** -0.5
        self.heads = n_heads
        hidden_dim = dim_keys * n_heads
        self.to_qkv = nn.Conv2d(n_channels, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Conv2d(hidden_dim, n_channels, 1)

    def forward(self, x):
        b, c, h, w = x.shape
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: einops.rearrange(t, "b (h c) x y -> b h c (x y)", h=self.heads), qkv
        )
        q = q * self.scale

        sim = torch.einsum("b h d i, b h d j -> b h i j", q, k)
        sim = sim - sim.amax(dim=-1, keepdim=True).detach()
        attention = sim.softmax(dim=-1)

        res = torch.einsum("b h i j, b h d j -> b h i d", attention, v)
        res = einops.rearrange(res, "b h (x y) d -> b (h d) x y", x=h, y=w)

        return self.to_out(res)


class LinearConvAttention(nn.Module):
    def __init__(self, n_channels: int, dim_keys: int = 32, n_heads: int = 2):
        """
        Efficient Attention (https://arxiv.org/abs/1812.01243), which instead of
        computing V (Q K.T) like in dot-product attention, computes Q (K.T V).
        This results in less complexity, O(d_k * d_v) instead of O(n²).

        Args:
            n_channels (int): Number of channels of the input feature maps
            dim_keys (int): Dimension of queries, keys, and values
            n_heads (int): Number of heads for attention
        """
        super().__init__()
        self.scale = dim_keys ** -0.5
        self.n_heads = n_heads
        hidden_dim = dim_keys * n_heads
        self.to_qkv = nn.Conv2d(n_channels, hidden_dim * 3, 1, bias=False)

        self.to_out = nn.Sequential(nn.Conv2d(hidden_dim, n_channels, 1),
                                    nn.GroupNorm(1, n_channels))

    def forward(self, x: torch.Tensor):
        b, c, h, w = x.shape
        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(
            lambda t: einops.rearrange(t, "b (h c) x y -> b h c (x y)", h=self.n_heads), qkv
        )

        q = q.softmax(dim=-2)
        k = k.softmax(dim=-1)

        q = q * self.scale
        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)

        res = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        res = einops.rearrange(res, "b h c (x y) -> b (h c) x y", h=self.n_heads, x=h, y=w)

        return self.to_out(res)


if __name__ == "__main__":
    ipt = torch.randn((4, 1024, 8, 8))

    attn = MultiHeadAttention(1024, dim=32, n_heads=4)
    out = attn(einops.rearrange(ipt, 'b c h w -> b (h w) c'))
    print("Multi-Head Attention")
    print("\tInput:", ipt.shape)
    print("\tOutput:", out.shape)

    attn = ConvAttention(1024, dim_keys=32, n_heads=2)
    out = attn(ipt)
    print("Conv Attention")
    print("\tInput:", ipt.shape)
    print("\tOutput:", out.shape)

    lin_attn = LinearConvAttention(1024, dim_keys=32, n_heads=2)
    out = lin_attn(ipt)
    print("Linear Conv Attention")
    print("\tInput:", ipt.shape)
    print("\tOutput:", out.shape)



import argparse

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

# DDP stuff
from torch.nn.parallel import DistributedDataParallel as DDP
import distributed_pytorch as dist


parser = argparse.ArgumentParser(description='PyTorch Multi-GPU Training')
parser.add_argument('--gpu', default=None, type=int, metavar='GPU',
                    help='Specify GPU for single GPU training. If not specified, it runs on all '
                         'CUDA_VISIBLE_DEVICES.')
parser.add_argument('--epochs', default=2, type=int, metavar='N',
                    help='Number of training epochs.')
parser.add_argument('--batch-size', default=8, type=int, metavar='N',
                    help='Batch size.')

# data
parser.add_argument('--n-classes', default=10, type=int, metavar='N',
                    help='Number of classes for fake dataset.')
parser.add_argument('--data-size', default=32, type=int, metavar='N',
                    help='Size of fake dataset.')
parser.add_argument('--image-size', default=64, type=int, metavar='N',
                    help='Size of input images.')


class DummyDataset(Dataset):
    def __init__(self, length, im_size, n_classes):
        self.len = length
        self.data = torch.randn(length, 3, im_size, im_size)
        self.labels = torch.randint(0, n_classes, (length,),
                                    generator=torch.Generator().manual_seed(0))

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]

    def __len__(self):
        return self.len


class DummyModel(nn.Module):
    def __init__(self, n_classes, in_channels=3):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, 64, kernel_size=7)
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.lin = nn.Linear(64, n_classes)

    def forward(self, x):
        x = self.conv(x)
        x = self.avg_pool(x)
        x = x.view(x.shape[0], -1)
        x = self.lin(x)

        return x


# Main workers ##################
def main_worker(gpu, world_size, args):
    args.gpu = gpu

    if args.distributed:
        dist.init_process_group(gpu, world_size)

    """ Data """
    dataset = DummyDataset(args.data_size, args.image_size, args.n_classes)
    sampler = dist.data_sampler(dataset, args.distributed, shuffle=False)
    loader = DataLoader(dataset, batch_size=args.batch_size,
                        shuffle=(sampler is None), sampler=sampler)

    """ Model """
    model = DummyModel(args.n_classes)

    # determine device
    if not torch.cuda.is_available():  # cpu
        device = torch.device('cpu')
    else:  # single or multi gpu
        device = torch.device(f'cuda:{args.gpu}')
    model.to(device)

    if args.distributed:
        model = DDP(model, device_ids=[args.gpu])

    """ Optimizer and Loss """
    # optimizer and loss
    optimizer = torch.optim.AdamW(model.parameters(), 0.0001)
    criterion = nn.CrossEntropyLoss().to(device)

    """ Run Epochs """
    for epoch in range(args.epochs):
        if dist.is_primary():
            print(f"------- Epoch {epoch + 1}")

        if args.distributed:
            sampler.set_epoch(epoch)

        # training
        train(model, loader, criterion, optimizer, device)

    # kill process group
    dist.cleanup()


def train(model, loader, criterion, optimizer, device):
    model.train()

    for it, (x, y) in enumerate(loader):
        x, y = x.to(device), y.to(device)

        y_hat = model(x)

        loss = criterion(y_hat, y)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        correct = torch.argmax(y_hat, dim=1).eq(y).sum()
        n = y.shape[0]

        # metrics per gpu/process
        print(f"Device: {x.device}"
              f"\n\tInput: \t{x.shape}"
              f"\n\tLoss:  \t{loss.cpu().item():.5f}"
              f"\n\tAcc:   \t{correct / n:.5f} ({correct}/{n})")

        # synchronize metrics across gpus/processes
        loss = dist.reduce(loss)
        correct = dist.reduce(correct)
        n = dist.reduce(torch.tensor(n).to(device))
        acc = correct / n

        # metrics over all gpus, printed only on the main process
        if dist.is_primary():
            print(f"Finish iteration {it}"
                  f" - acc: {acc.cpu().item():.4f} ({correct}/{n})"
                  f" - loss: {loss.cpu().item():.4f}")


if __name__ == "__main__":
    # only run once
    parsed_args = parser.parse_args()
    for name, val in vars(parsed_args).items():
        print("{:<12}: {}".format(name, val))

    # start different processes, if multi-gpu is available
    # otherwise, it just starts the main_worker once
    dist.launch(main_worker, parsed_args)



import torch
import torch.nn as nn


class TimeEmbedding(nn.Module):
    def __init__(self, time_emb_dim: int, pos_emb_dim: int, max_len: int = 5000):
        """
        Time embedding for time step t. First, t is embedded using a fixed sinusoidal positional
        embedding, as described in "Attention Is All You Need" (https://arxiv.org/abs/1706.03762),
        followed by a two layer MLP.

        Args:
            time_emb_dim: Dimension of final time embedding
            pos_emb_dim: Embedding dimension for the fixed sinusoidal positional embedding
            max_len: Maximum number of time steps (default: 5000)
        """
        super().__init__()

        self.pos_emb_dim = pos_emb_dim
        self.time_emb_dim = time_emb_dim
        self.max_len = max_len

        # fixed sinusoidal positional embedding
        assert self.pos_emb_dim % 2 == 0, "Embedding dim must be a multiple of 2!"
        pos = torch.arange(0, max_len).float().unsqueeze(1)
        _2i = torch.arange(0, self.pos_emb_dim, 2).float()
        pos_embedding = torch.zeros(self.max_len, self.pos_emb_dim)
        pos_embedding[:, 0::2] = torch.sin(pos / (10000 ** (_2i / self.pos_emb_dim)))
        pos_embedding[:, 1::2] = torch.cos(pos / (10000 ** (_2i / self.pos_emb_dim)))
        self.register_buffer('pos_embedding', pos_embedding, persistent=True)

        # MLP for time embedding
        self.mlp = nn.Sequential(
            nn.Linear(self.pos_emb_dim, self.time_emb_dim),
            nn.SiLU(),
            nn.Linear(self.time_emb_dim, self.time_emb_dim)
        )

    def forward(self, t: torch.Tensor):
        """
        Embeds a time step t with a fixed sinusoidal positional encoding, and then
        transforms it with an MLP.

        Args:
            t (torch.Tensor): Batch of time steps with shape [bs,]

        Returns:
            t_emb (torch.Tensor): Time embeddings for the respective time indices
        """
        t_pos_emb = torch.index_select(self.pos_embedding, 0, t)
        t_emb = self.mlp(t_pos_emb)

        return t_emb


if __name__ == "__main__":
    emb = TimeEmbedding(time_emb_dim=16, pos_emb_dim=4, max_len=10)
    time = torch.randint(0, 10, (4,))
    time_emb = emb.forward(time)

    print("Fixed positional embedding:\n", emb.pos_embedding)
    print("Time:", time)                        # eg [2, 5, 7, 9]
    print("Time embedding:", time_emb.shape)    # [4, 16]



from tajer.nn.attention import MultiHeadAttention, LinearConvAttention, ConvAttention
from tajer.nn.cbam import CBAM
from tajer.nn.dsconv import DepthwiseSeparableConv2D
from tajer.nn.residual import ResidualBlock
from tajer.nn.time_embedding import TimeEmbedding



import os
import time
import math
import yaml
import torch
import einops
import datetime

import numpy as np
from PIL import Image
import torchvision.transforms as T


def exists(x):
    return x is not None


def is_odd(n):
    return (n % 2) == 1


def default(val, d):
    if exists(val):
        return val
    return d() if callable(d) else d


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


# size of pytorch tensors

def get_tensor_size(tensor: torch.Tensor, return_bytes=False):
    size = tensor.element_size() * tensor.nelement()
    if return_bytes:
        return size
    return convert_size(size)


# time stuff

def timer(start, end):
    hours, rem = divmod(end - start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)


def timing():
    return datetime.datetime.now().strftime(f'%H:%M:%S.%f')[:-3]


# pytorch model parameter count

def count_parameters(model, return_int: bool = False):
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    if return_int:
        return n_params

    return f'{n_params:,}'


# namespace to access and set attributes with dot notation

class ConfigNamespace(dict):
    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            return None

    def __setattr__(self, attr, value):
        self[attr] = value


# overwrite config file information with arguments if arguments are set

def parse_options(parser):
    args = parser.parse_args()
    cfg = yaml.load(open(args.cfg, 'r'), Loader=yaml.Loader)

    # get data config
    data_cfg_file = args.data_cfg if exists(args.data_cfg) else cfg['data_cfg']
    data_cfg = yaml.load(open(data_cfg_file, 'r'), Loader=yaml.Loader)

    # set all params from config file
    for key, value in cfg.items():
        setattr(args, key, value)

    # overwrite with args
    overwrite_args = parser.parse_args()
    for key, value in vars(overwrite_args).items():
        if exists(value):
            setattr(args, key, value)

    # add data config as namespace object
    args.data = ConfigNamespace()
    for key, value in data_cfg.items():
        args.data[key] = value

    return args


# PyTorch average meter to accumulate and aggregate values

class AverageMeter:
    def __init__(self):
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0

    def update(self, val, n=1):
        val = self._handle_value(val)
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def reset(self):
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0

    @staticmethod
    def _handle_value(value):
        if isinstance(value, torch.Tensor):
            return value.item()
        return value


# Image and video stuff


def video_tensor_to_gif(tensor, path, duration=120, loop=0, optimize=True):
    # tensor of shape (channels, frames, height, width) -> gif
    images = map(T.ToPILImage(), tensor.unbind(dim=1))
    first_img, *rest_imgs = images
    first_img.save(path, save_all=True, append_images=rest_imgs, duration=duration, loop=loop, optimize=optimize)
    return images


def get_original_reconstruction_image(x: torch.Tensor, x_hat: torch.Tensor, n_ims: int = 8):
    """
    Returns pillow image of original and reconstruction images. Top row are originals, bottom
    row are reconstructions. Faster than creating a figure.
    Args:
        x: Original image tensor (in range -1 to 1)
        x_hat: Reconstructed image tensor
        n_ims: Number of images of that batch to be plotted
    Returns:
        ims: Numpy array in shape [h, w, 3] with top row being originals and
            bottom row being reconstructions.
    """
    bs, c, h, w = x.shape

    n_ims = n_ims if n_ims <= bs else bs

    x = x[:n_ims, ...].detach().cpu().numpy()
    x_hat = x_hat[:n_ims, ...].detach().cpu().numpy()

    x = np.transpose(x, (0, 2, 3, 1))
    x_hat = np.transpose(x_hat, (0, 2, 3, 1))

    x = einops.rearrange(x, 'b h w c -> h (b w) c')
    x_hat = einops.rearrange(x_hat, 'b h w c -> h (b w) c')

    ims = np.concatenate((x, x_hat), axis=0)

    ims = (ims * 127.5 + 127.5).astype(np.uint8)

    return ims


if __name__ == "__main__":
    print(f"{'timing()':<18}: {timing()}")
    print(f"{'timer(s, e)':<18}: {timer(time.time(), time.time() + 900)}")
    
    print(f"{'getsize()':<18}: {get_tensor_size(torch.randn((4096, 1024)))}")
    print(f"{'count_parameters()':<18}: {count_parameters(torch.nn.Conv2d(128, 256, 5))}")



# MIT License Copyright (c) 2022 joh-fischer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import torch

import torch.multiprocessing as mp
from torch import distributed as dist
from torch.utils.data.distributed import DistributedSampler


# launch
def launch(worker_fn, args):
    world_size = torch.cuda.device_count()

    args.single_gpu = world_size > 0 and args.gpu is not None
    args.distributed = world_size > 1 and not args.single_gpu

    if args.distributed:
        if args.distributed and "CUDA_VISIBLE_DEVICES" not in os.environ:
            raise ValueError("GPUs not specified. Please set CUDA_VISIBLE_DEVICES before"
                             " running the script or specify a single GPU via '--gpu'.")

        os.environ["NCCL_P2P_DISABLE"] = "1"
        mp.spawn(worker_fn, args=(world_size, args),
                 nprocs=world_size, join=True)
    elif args.single_gpu:
        worker_fn(args.gpu, world_size, args)
    else:
        worker_fn(0, world_size, args)


# distributed trainings functions
def init_process_group(rank, world_size, port='12352', backend='nccl'):
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = port

    dist.init_process_group(backend, init_method="env://",
                            rank=rank, world_size=world_size)


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def cleanup():
    if is_dist_avail_and_initialized():
        dist.destroy_process_group()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0

    return dist.get_rank()


def is_primary():
    return get_rank() == 0


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


# data loading stuff
def data_sampler(dataset, distributed, shuffle):
    if distributed:
        return DistributedSampler(dataset, shuffle=shuffle)

    return None


# synchronization functions
def all_reduce(tensor, op='sum'):
    world_size = get_world_size()

    if world_size == 1:
        return tensor

    if op == 'sum':
        dist.all_reduce(tensor, op=dist.ReduceOp.SUM)
    elif op == 'avg':
        dist.all_reduce(tensor, op=dist.ReduceOp.SUM)
        tensor /= get_world_size()
    else:
        raise ValueError(f'"{op}" is an invalid reduce operation!')

    return tensor


def reduce(tensor, op=dist.ReduceOp.SUM):
    world_size = get_world_size()

    if world_size == 1:
        return tensor

    dist.reduce(tensor, dst=0, op=op)

    return tensor


def gather(data):
    world_size = get_world_size()

    if world_size == 1:
        return [data]

    output_list = [torch.zeros_like(data) for _ in range(world_size)]

    if is_primary():
        dist.gather(data, gather_list=output_list)
    else:
        dist.gather(data)

    return output_list


def synchronize():
    world_size = get_world_size()

    if world_size == 1:
        return

    dist.barrier()



# Adapted from https://github.com/HansBambel/SmaAt-UNet
import torch
import torch.nn as nn


class DepthwiseSeparableConv2D(nn.Module):
    def __init__(self, in_channels: int, output_channels: int, kernel_size: int,
                 padding: int = 0, kernels_per_layer: int = 1):
        super(DepthwiseSeparableConv2D, self).__init__()
        self.depth_wise = nn.Conv2d(in_channels, in_channels * kernels_per_layer,
                                    kernel_size=kernel_size, padding=padding, groups=in_channels)
        self.point_wise = nn.Conv2d(in_channels * kernels_per_layer, output_channels, kernel_size=1)

    def forward(self, x: torch.Tensor):
        x = self.depth_wise(x)
        x = self.point_wise(x)

        return x


if __name__ == "__main__":
    ipt = torch.randn((8, 3, 64, 64))
    dsc = DepthwiseSeparableConv2D(3, 16, 3, padding=1)

    print("input:", ipt.shape)          # torch.Size([8, 3, 64, 64])
    print("output:", dsc(ipt).shape)    # torch.Size([8, 16, 64, 64])



import os
import torch
import pandas as pd
from logger import Logger


experiment_name = 'model1'

logger = Logger('./logs',
                # create log-folder: './logs/model1/22-07-07_121028'
                experiment_name, timestamp=True,
                # include tensorboard SummaryWriter
                tensorboard=True)

logger.log_hparams({'lr': 1e-4,
                    'optimizer': 'Adam'})

for epoch in range(2):
    logger.init_epoch(epoch)     # initialize epoch to aggregate values

    # training
    for step in range(4):
        logger.log_metrics({'loss': torch.rand(1), 'acc': torch.rand(1)},
                           phase='train', aggregate=True)

    # write to tensorboard
    logger.tensorboard.add_scalar('train/loss', logger.epoch['loss'].avg)

    # validation simulation
    for step in range(2):
        logger.log_metrics({'val_loss': torch.rand(1)},
                           phase='val', aggregate=True)

        print('Running average:', logger.epoch['val_loss'].avg)
        print('Running sum:', logger.epoch['val_loss'].sum)

logger.save()

metrics_path = os.path.join(logger.log_dir, 'metrics.csv')
print(pd.read_csv(metrics_path).head(20))



import torch
import torch.nn as nn


class ResidualBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, time_emb_dim: int = None, n_groups: int = 8):
        """
        Residual block with time conditioning.

        Args:
            in_channels: Input channels to residual block
            out_channels: Output channels of residual block
            time_emb_dim: Dimension of time embedding
            n_groups: Number of groups for group normalization
        """
        super().__init__()

        self.block1 = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.GroupNorm(n_groups, out_channels),
            nn.SiLU()
        )

        self.block2 = nn.Sequential(
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.GroupNorm(n_groups, out_channels),
            nn.SiLU()
        )

        if in_channels != out_channels:
            self.shortcut = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        else:
            self.shortcut = nn.Identity()

        self.time_emb = nn.Linear(time_emb_dim, out_channels) if time_emb_dim is not None else None

    def forward(self, x: torch.Tensor, t: torch.Tensor = None):
        identity = self.shortcut(x)

        x = self.block1(x)

        # condition with time if required
        if self.time_emb is not None:
            t = self.time_emb(t)        # [bs, out_channels]
            x += t[:, :, None, None]

        x = self.block2(x)

        return x + identity


if __name__ == "__main__":
    bs = 8
    t_channels = 16

    ipt = torch.randn((bs, 64, 32, 32))
    time = torch.randn((bs, t_channels))

    res = ResidualBlock(64, 128, t_channels)
    out = res(ipt, time)

    print("Input:", ipt.shape)
    print("Time:", time.shape)
    print("Output:", out.shape)



import torch
import torch.nn as nn


# adapted from https://github.com/openai/improved-diffusion/blob/main/improved_diffusion/nn.py

def checkpoint(func, inputs, params, flag):
    """
    Evaluate a function without caching intermediate activations, allowing for
    reduced memory at the expense of extra compute in the backward pass.

    :param func: the function to evaluate.
    :param inputs: the argument sequence to pass to `func`.
    :param params: a sequence of parameters `func` depends on but does not
                   explicitly take as arguments.
    :param flag: if False, disable gradient checkpointing.
    """
    if flag:
        args = tuple(inputs) + tuple(params)
        return CheckpointFunction.apply(func, len(inputs), *args)
    else:
        return func(*inputs)


class CheckpointFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, run_function, length, *args):
        ctx.run_function = run_function
        ctx.input_tensors = list(args[:length])
        ctx.input_params = list(args[length:])
        with torch.no_grad():
            output_tensors = ctx.run_function(*ctx.input_tensors)
        return output_tensors

    @staticmethod
    def backward(ctx, *output_grads):
        ctx.input_tensors = [x.detach().requires_grad_(True) for x in ctx.input_tensors]
        with torch.enable_grad():
            # Fixes a bug where the first op in run_function modifies the
            # Tensor storage in place, which is not allowed for detach()'d
            # Tensors.
            shallow_copies = [x.view_as(x) for x in ctx.input_tensors]
            output_tensors = ctx.run_function(*shallow_copies)
        input_grads = torch.autograd.grad(
            output_tensors,
            ctx.input_tensors + ctx.input_params,
            output_grads,
            allow_unused=True,
        )
        del ctx.input_tensors
        del ctx.input_params
        del output_tensors
        return (None, None) + input_grads


# Dummy class just for demonstration purposes
class Dummy(nn.Module):
    def __init__(self, in_ch, out_ch, use_checkpoint):
        super().__init__()
        self.use_checkpoint = use_checkpoint
        self.net = nn.Linear(in_ch, out_ch)

    def _forward(self, x: torch.Tensor, emb=None):
        out = self.net(x)
        return out

    def forward(self, x: torch.Tensor, emb=None):
        return checkpoint(
            self._forward, (x, emb), self.parameters(), self.use_checkpoint
        )


if __name__ == "__main__":
    model = Dummy(in_ch=128, out_ch=256, use_checkpoint=True)

    ipt = torch.randn((32, 128))
    out = model(ipt)
    print("Input:", ipt.shape)
    print("Output:", out.shape)



import torch
import torch.nn as nn


# Adapted from: https://github.com/luuuyi/CBAM.PyTorch/blob/master/model/resnet_cbam.py
class ChannelAttention(nn.Module):
    def __init__(self, channels: int, reduction_ratio: int = 16):
        """
        Compute channel attention (average- and max-pooled features) to model
        inter-channel relationship ('what' to focus). This squeezes the spatial
        dimension of the input feature map, s.t. the input CxHxW will be mapped
        to Cx1x1. The argument `hidden_ch` can be used to create a bottleneck
        in the mlp (https://arxiv.org/abs/1807.06521).
        Args:
            channels: Number of channels of the feature maps.
            reduction_ratio: Reduction ratio for bottleneck.
        """
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.mlp = nn.Sequential(nn.Conv2d(channels, channels // reduction_ratio, 1, bias=False),
                                 nn.ReLU(),
                                 nn.Conv2d(channels // reduction_ratio, channels, 1, bias=False))

    def forward(self, x: torch.Tensor):
        avg_out = self.mlp(self.avg_pool(x))
        max_out = self.mlp(self.max_pool(x))
        out = avg_out + max_out

        return torch.sigmoid(out)


class SpatialAttention(nn.Module):
    def __init__(self, kernel_size: int = 7):
        """
        Compute spatial attention (average- and max-pooled features) to model
        attention along the spatial dimensions ('where' to focus). This squeezes
        the channel dimension of the input feature map, s.t. the input CxHxW will
        be mapped to 1xHxW (paper: https://arxiv.org/abs/1807.06521).
        Args:
            kernel_size: Kernel size for the convolution.
        """
        super().__init__()
        self.conv = nn.Conv2d(2, 1, kernel_size, padding=kernel_size // 2, bias=False)

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        out = self.conv(x)

        return torch.sigmoid(out)


class CBAM(nn.Module):
    def __init__(self, in_channels: int, reduction_ratio: int = 16, kernel_size: int = 7):
        super().__init__()
        self.channel_attn = ChannelAttention(in_channels, reduction_ratio)
        self.spatial_attn = SpatialAttention(kernel_size)

    def forward(self, x: torch.Tensor):
        x *= self.channel_attn(x)
        x *= self.spatial_attn(x)

        return x


if __name__ == "__main__":
    ipt = torch.randn((8, 32, 64, 64))
    print("Input:", ipt.shape)              # torch.Size([8, 32, 64, 64])

    c_attn = ChannelAttention(32)
    print("Channel:", c_attn(ipt).shape)    # torch.Size([8, 32, 1, 1])

    s_attn = SpatialAttention()
    print("Spatial:", s_attn(ipt).shape)    # torch.Size([8, 1, 64, 64])

    cbam = CBAM(32)
    print("CBAM:", cbam(ipt).shape)         # torch.Size([8, 32, 64, 64])



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
install_requires = (this_directory / "requirements.txt").read_text().splitlines()
long_description = (this_directory / "README.md").read_text()

setup(
    name="tajer",
    version="1.2.2",
    packages=find_packages(),
    url="https://github.com/joh-fischer/tajer",
    license="MIT License",
    author="Johannes S. Fischer",
    description="Collection of useful (PyTorch) functions.",
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
)
